# accelerator.py

'''Create an accelerator.
'''

import numpy as np
from scipy import interpolate

__all__ = ['Element',
           'Accelerator',
           'read_elements']


class Element:
    '''Sets one accelerator element.

    Sets one accelerator element with parameters:
    z0 [m] --- element position,
    max_field --- maximum field, 
    file_name --- field profile,
    name --- unique element name.
    '''
    def __init__(self,
                 z0: float,
                 max_field: float,
                 file_name: str,
                 name: str):
        self.z0 = z0
        self.max_field = max_field
        self.file_name = file_name
        self.name = name


field_files = {} # buffer for field files

def read_elements(z:np.arange,
                  beamline: dict) -> interpolate.interp1d:
    '''Sews elements into a function of z.

    Sews elements into a function of z with parameters:
    beamline --- set of accelerator elements,
    z [m] --- coordinate.
    '''
    global field_files
    F = 0
    if not beamline:
        z_data = [i/1000 for i in range(1000)]
        F_data = [0 for i in range(1000)]
        f = interpolate.interp1d(
            z_data, F_data,
            fill_value=(0, 0), bounds_error=False
        )
        F = F + f(z)
    else:
        for element in beamline.values():
            if not (element.file_name in field_files):
                field_files[element.file_name] = np.loadtxt(element.file_name)
            M = field_files[element.file_name]
            z_data = M[:,0]
            F_data = M[:,1]
            f = interpolate.interp1d(
                element.z0+z_data, element.max_field*F_data,
                fill_value=(0, 0), bounds_error=False
            )
            F = F + f(z)
    F = interpolate.interp1d(z, F, fill_value=(0, 0), bounds_error=False)
    return F

        
class Accelerator:
    '''Create an accelerator.
    
    Create an accelerator with parameters:
    start [m] --- the beginning of the accelerator,
    stop [m] --- end of the accelerator,
    step [m] --- step method along the accelerator.
    
    Can add a solenoids, quadrupoles and accelerating modules.
    '''
    def __init__(self, 
                 start: float,
                 stop: float,
                 step: float):
        self.start = start
        self.stop = stop
        self.step = step
        self.parameter = np.arange(start, stop, step)
        self.Bz_beamline = {}
        self.Ez_beamline = {}
        self.Gz_beamline = {}
        self.Bz = interpolate.interp1d
        self.Ez = interpolate.interp1d
        self.Gz = interpolate.interp1d
    
    def add_solenoid(self,
                     name: str,
                     center: float,
                     max_field: float,
                     file_name: str ) -> None:
        '''
        Creates a solenoid in the accelerator.
        
        Creates a solenoid in the accelerator with parameters:
        name --- solenoid's id,
        centert [m] --- solenoid's center,
        max field [T] --- solenoid's maximum field,
        file name --- experimental profile of the Bz field,
        '''
        self.Bz_beamline[name] = Element(center, max_field, file_name, name)
    
    def add_accel(self,
                     name: str,
                     center: float,
                     max_field: float,
                     file_name: str ) -> None:
        '''
        Creates an accelerating module in the accelerator.
        
        Creates an accelerating module in the accelerator with parameters:
        name --- accelerating module's id,
        centert [m] --- accelerating module's center,
        max field [MV/m] --- accelerating module's maximum field,
        file name --- experimental profile of the Ez field,
        '''
        self.Ez_beamline[name] = Element(center, max_field, file_name, name)
    
    def add_quadrupole(self,
                     name: str,
                     center: float,
                     max_field: float,
                     file_name: str ) -> None:
        '''
        Creates a quadrupole in the accelerator.
        
        Creates a quadrupole in the accelerator with parameters:
        name --- quadrupole's id,
        centert [m] --- quadrupole's center,
        max field [T/m] --- quadrupole's maximum field,
        file name --- experimental profile of the Gz field,
        '''
        self.Gz_beamline[name] = Element(center, max_field, file_name, name)
    
    def delete_solenoid(self,
                     name: str='all') -> None:
        '''
        Delete a solenoid in the accelerator.
        
        Delete a solenoid in the accelerator with parameters:
        name --- solenoid's id
        
        *if there is no name that will be removed all
        '''
        if name == 'all':
            self.Bz_beamline = {}
        else:
            self.Bz_beamline.pop(name)
    
    def delete_accel(self,
                     name: str='all') -> None:
        '''
        Delete a accelerating module in the accelerator.
        
        Delete a accelerating module in the accelerator with parameters:
        name --- quadrupole's id
        
        *if there is no name that will be removed all
        '''
        if name == 'all':
            self.Ez_beamline = {}
        else:
            self.Ez_beamline.pop(name)
    
    def delete_quadrupole(self,
                     name: str='all') -> None:
        '''
        Delete a quadrupole in the accelerator.
        
        Delete a quadrupole in the accelerator with parameters:
        name --- quadrupole's id
        
        *if there is no name that will be removed all
        '''
        if name == 'all':
            self.Gz_beamline = {}
        else:
            self.Gz_beamline.pop(name)
    
    
    def compile(self) -> None:
        '''Compilation of the accelerator.
        '''
        self.Bz = read_elements(self.parameter, self.Bz_beamline)
        self.Ez = read_elements(self.parameter, self.Ez_beamline)
        self.Gz = read_elements(self.parameter, self.Gz_beamline)
        
        print('Accelerator compiled.')
     
    def __str__(self):
        string = 'Accelerator structure.\n'
        string += '\tSolenoids:\n'
        for element in self.Bz_beamline.values():
            string +="\t[ %.5f m, %.5f T, '%s', '%s'],\n" % (element.z0, element.max_field, element.file_name, element.name)
        string += '\tAccelerating modules:\n'
        for element in self.Ez_beamline.values():
            string +="\t[ %.5f m, %.5f Mv/m, '%s', '%s'],\n" % (element.z0, element.max_field, element.file_name, element.name)
        string += '\tQuadrupoles:\n'
        for element in self.Gz_beamline.values():
            string +="\t[ %.5f m, %.5f T/m, '%s', '%s'],\n" % (element.z0, element.max_field, element.file_name, element.name)
        return string
    
    add_sol = add_new_solenoid = add_solenoid
    add_acc = add_new_accel = add_accel
    add_quad = add_new_quadrupole = add_quadrupole
    
    del_sol = del_solenoid = delete_solenoid
    del_acc = del_accel = delete_accel
    del_quad = del_quadrupole = delete_quadrupole



# solver.py
'''Simulation of the envelope beam in the accelerator.
'''
import numpy as np

from .beam import *
from .accelerator import *
from scipy import interpolate, integrate
from scipy.integrate import odeint
from scipy.misc import derivative


__all__ = ['Sim',
           'Simulation']

class Simulation:
    '''Simulation of the envelope beam in the accelerator.
    '''
    def __init__(self,
             beam,
             accelerator):
        
        self.mass_rest_electron = 0.511
        self.clight = 299792458
        self.alfven_current = 17000
        self.rtol = 0.001
        
        self.current = beam.current
        self.energy = beam.energy
        self.radius_x = beam.radius_x
        self.radius_y = beam.radius_y
        self.angular_x = beam.angular_x
        self.angular_y = beam.angular_y
        self.normalized_emittans_x = beam.normalized_emittans_x
        self.normalized_emittans_y = beam.normalized_emittans_y

        self.start = accelerator.start
        self.stop = accelerator.stop
        self.step = accelerator.step
        self.parameter = accelerator.parameter
        self.Bz = accelerator.Bz
        self.Ez = accelerator.Ez
        self.Gz = accelerator.Gz
        
        self.gamma_0 = self.energy/self.mass_rest_electron + 1
        self.gamma = self.gamma_0 + integrate.cumtrapz(-self.Ez(self.parameter)/self.mass_rest_electron, self.parameter)
        self.gamma = interpolate.interp1d(self.parameter[1:], self.gamma, fill_value=(self.gamma[0], self.gamma[-1]), bounds_error=False)
        
        self.envelope_x = []
        self.envelope_y = []
    
    def track(self):
        '''Tracking!
        '''
        
        def dXdz(X:list,
                 z:np.arange,
                 dz:float=self.step,
                 gamma:interpolate.interp1d=self.gamma,
                 Bz:interpolate.interp1d=self.Bz,
                 Ez:interpolate.interp1d=self.Ez, 
                 Gz:interpolate.interp1d=self.Gz,
                 beam_current:float=self.current,
                 emitt_n_x:float=self.normalized_emittans_x,
                 emitt_n_y:float=self.normalized_emittans_y,
                 c:float=self.clight,
                 mc:float=self.mass_rest_electron,
                 alfven_current:float=self.alfven_current) ->list:
            '''Located derivative for further integration.
            '''
            
            x, sigma_x, y, sigma_y = X[0], X[1], X[2], X[3] 
    
            g = gamma(z)
            dgdz = Ez(z) / mc
            d2gdz2 = derivative(lambda z:Ez(z),z, dz) / mc

            beta = np.sqrt(1 - 1 / (g * g))

            K_s = ( c * Bz(z) / (2*0.511e6) / (g * beta))**2
            K_q = ( c * Gz(z) / (0.511e6)) / (g * beta)
            K_x = K_s + K_q
            K_y = K_s - K_q

            emitt_x = emitt_n_x / (g * beta)
            emitt_y = emitt_n_y / (g * beta)

            P = 2 * beam_current / (alfven_current * beta * beta * beta * g * g * g)

            dxdz = 2 * P / (sigma_x + sigma_y) + emitt_x * emitt_x / (sigma_x * sigma_x * sigma_x) - K_x * sigma_x - \
                   dgdz * x / (beta * beta * g) - d2gdz2 * sigma_x/(2 * beta * beta * g)
            dsigma_xdz = x

            dydz = 2 * P/(sigma_x + sigma_y) + emitt_y * emitt_y / (sigma_y * sigma_y * sigma_y) - K_y * sigma_y - \
                   dgdz * y/(beta * beta * g) - d2gdz2 * sigma_y/(2 * beta * beta * g)
            dsigma_ydz = y

            return [dxdz, dsigma_xdz, dydz, dsigma_ydz]

        X0 = [self.radius_x, self.angular_x, self.radius_y, self.angular_y]
        self.envelope_x = odeint(dXdz, X0, self.parameter, rtol=self.rtol)[0:,1]
        self.envelope_y = odeint(dXdz, X0, self.parameter, rtol=self.rtol)[0:,3]
        
        print('Tracking is completed.')
        
Sim = Simulation


# beam.py

'''Creating an electron beam.
'''

__all__ = ['Beam']

class Beam:
    '''Creating an electron beam.
    
    Creating an elliptical electron beam with parameters:
    current [A],
    energy [MeV],
    radius x [m],
    radius y [m],
    angular x [rad],
    angular y [rad],
    normalized emittans x [m*rad], 
    normalized emittans y [m*rad].
    '''
    
    def __init__(self,
                 current: float=.0e0,
                 energy: float=.0e0,
                 radius: float=.0e0,
                 radius_x: float=.0e0,
                 radius_y: float=.0e0,
                 angular: float=.0e0,
                 angular_x: float=.0e0,
                 angular_y: float=.0e0,
                 normalized_emittans: float=.0e0,
                 normalized_emittans_x: float=.0e0,
                 normalized_emittans_y: float=.0e0):
        
        self.current = current
        self.energy = energy
        self.radius_x = radius
        self.radius_y = radius
        self.angular_x = angular
        self.angular_y = angular
        self.normalized_emittans_x = normalized_emittans
        self.normalized_emittans_y = normalized_emittans
        if (radius_x or radius_y):
            self.radius_x = radius_x
            self.radius_y = radius_y   
        if (angular_x or angular_y):
            self.angular_x = angular_x
            self.angular_y = angular_y
        if (normalized_emittans_x or normalized_emittans_y):
            self.normalized_emittans_x = normalized_emittans_x
            self.normalized_emittans_y = normalized_emittans_y
            
    
    def __str__(self):
            return 'Beam parameters:' + '\n' \
                    +'\tCurrent\t%0.1f A'%(self.current) + '\n' \
                    +'\tEnergy\t%0.1f MeV'%(self.energy) + '\n' \
                    +'\tRadius x\t%0.1f mm'%(self.radius_x*1e3) + '\n' \
                    +'\tRadius y\t%0.1f mm'%(self.radius_y*1e3) + '\n' \
                    +'\tAngular x\t%0.1f mrad'%(self.angular_x*1e3) + '\n' \
                    +'\tAngular y\t%0.1f mrad'%(self.angular_y*1e3) + '\n' \
                    +'\tNormalized emittans x\t%0.1f mm*mrad'%(self.normalized_emittans_x*1e6) + '\n' \
                    +'\tNormalized emittans y\t%0.1f mm*mrad'%(self.normalized_emittans_y*1e6) + '\n'